import java.awt.Dimension;
import java.io.PrintStream;
import java.nio.*;
import java.util.ArrayList;

import processing.core.*;
import processing.opengl.*;

import controlP5.ControlEvent;
import controlP5.ControlP5;

import peasy.*;
import peasy.org.apache.commons.math.geometry.Vector3D;

import toxi.geom.*;
import toxi.math.MathUtils;
import toxi.processing.*;

import javax.media.opengl.*;
import javax.media.opengl.glu.*;

import codeanticode.gsvideo.*;

/**
 * 
 * @author mannytan
 * @name NudiBranch
 * @date Dec 11, 2010
 * 
 */
public class NudiBranch extends PApplet {

	GL gl;
	GLU glu;

	// create Guides
	Guides guides;

	// camera for motion/zoomin
	PeasyCam cam;

	int totalSegments = 62;

	Segment firstSegment;

	ArrayList<Segment> segmentList = new ArrayList();

	public static ToxiclibsSupport gfx;

	// mouse control of nudibranch, not used in this version
	private Vec3D mouseVec;

	// speed of nudibranch motion
	private float speeder = 0;

	// controls
	private Boolean isControlsVisible = false;
	private Controls controls;

	private About about;

	// audio
	private Audio audio;

	// movie recorded
	GSMovieMaker mm;

	int fps = 30;
	int resetCount = 1;

	//*
	static public void main(String args[]) {
		PApplet.main(new String[] { "--present", "NudiBranch" });
		// PApplet.main(new String[] { "NudiBranch" });
	}

	//*/

	public void setup() {
		// size(screen.width, screen.height, OPENGL);
		size(1200, 800, OPENGL);
		// hint(DISABLE_DEPTH_TEST);
		// hint(ENABLE_DEPTH_SORT);
		// this.setSize(new Dimension(width, height)); // resize & set bounds of the PApplet
		// this.setBounds(0, 0, width, height);
		frameRate(fps);
		background(15);

		// Create Camera
		cam = new PeasyCam(this, 1000);
		cam.lookAt(0, 0, 0, 0);

		// Create ToxiclibsSupport
		gfx = new ToxiclibsSupport(this);

		// Create controls
		controls = new Controls(this);
		controls.createControls();
		controls.setAsEndSliders();


		// Create controls
		about = new About(this);
		about.createControls();

		
		// Create audio
		audio = new Audio(this);
		audio.setTrack("recognizer.mp3");
		// audio.setTrack("jingle.mp3");
		// audio.play();

		// Create Guides
		guides = new Guides(this);

		// Create first segment data
		firstSegment = new Segment(this);
		firstSegment.seed = 0;
		firstSegment.seedNormal = 0.0f;
		firstSegment.seedNormalReversed = 1.0f - firstSegment.seedNormal;
		firstSegment.parentSegment = firstSegment;

		// Create Segments
		for (int i = 0; i < totalSegments; i++) {
			Segment segment = new Segment(this);
			segment.seed = (i + 1);
			segment.seedNormal = (float) (i + 1) / (float) totalSegments;
			segment.seedNormalReversed = 1.0f - segment.seedNormal;

			// if first segment, make parent itself
			if (i == 0) {
				segment.parentSegment = firstSegment;
			} else {
				segment.parentSegment = segmentList.get(segmentList.size() - 1);
			}

			segment.randomizeCurrentControls();
			segment.setPreviousControls();
			segmentList.add(segment);
		}

		// Create 3d mouse data
		gl = ((PGraphicsOpenGL) g).gl;
		glu = ((PGraphicsOpenGL) g).glu;
		mouseVec = new Ray3D();

		// mm = new GSMovieMaker(this, width, height, "drawing.ogg", GSMovieMaker.THEORA, GSMovieMaker.BEST, fps);
		// mm.start();
	}

	public void draw() {
		if (frameCount % 300 == 0) {
			reset();
		}

		background(15);
		this.rotateY(frameCount * .01f);

		// update controls
		controls.updateParams();

		// update audio
		audio.updateAudioParams();

		// update control params
		updateControlParams();

		// camera HUD
		cam.beginHUD();
		if (controls.isEnvironmentVisible) {
			guides.draw2DScope();
		}

		// hide the controls
		about.draw();

		if (isControlsVisible) {
			controls.draw();
			about.draw();

		}
		cam.endHUD();

		// add lights
		lights();

		// draw guides
		if (controls.isGuideVisible) {
			guides.draw3DAxes(400.0f);
		}
		if (controls.isEnvironmentVisible) {
			guides.draw3DBox(400.0f);
		}

		if (controls.isMouseActive) {
			// update pointer with mouse position
			float[] mousePosList = getMouse3D();
			mouseVec.x += (mousePosList[0] - mouseVec.x) * .05f;
			mouseVec.y += (mousePosList[1] - mouseVec.y) * .05f;
			mouseVec.z += (mousePosList[2] - mouseVec.z) * .05f;
			mouseVec.normalizeTo(1000.0f);

			/*
			// limiters
			if (mouseVec.z < -1000) {
				mouseVec.z = -1000;
			}
			if (mouseVec.z > 1000) {
				mouseVec.z = 1000;
			}
			*/

			// */
		} else {
			//*
			// update pointer with sinusoidal motion
			speeder += .025f * controls.rootSpeed;
			mouseVec.x = sin(MathUtils.PI * speeder * .5f) * controls.rootRadius + 0.0f;
			mouseVec.y = cos(MathUtils.PI * speeder) * controls.rootRadius + 0.0f;
			mouseVec.z = sin(MathUtils.PI * speeder) * controls.rootRadius + 0.0f;

			// */
		}
		// first segment
		firstSegment.current.set(mouseVec);
		firstSegment.previous.set(mouseVec);
		firstSegment.parse();

		Segment segment;
		segment = segmentList.get(0);
		segment.parse();

		int startId = 2;
		for (int i = 0; i < totalSegments; i++) {
			segment = segmentList.get(i);
			segment.parse();
		}

		if (controls.isFillMeshVisible) {
			for (int i = startId; i < totalSegments; i++) {
				segment = segmentList.get(i);
				segment.drawFillMesh(gfx);
				segment.drawLineTrailMesh(gfx);
				segment.drawLineHairMesh(gfx);
			}
		}

		if (controls.isDotMeshVisible) {
			for (int i = startId; i < totalSegments; i++) {
				segment = segmentList.get(i);
				segment.drawDotMesh(gfx);
			}
		}
		if (controls.isLineTrailVisible) {
			for (int i = startId; i < totalSegments; i++) {
				segment = segmentList.get(i);
				segment.drawLineTrailMesh(gfx);
			}
		}
		if (controls.isLineHairVisible) {
			for (int i = startId; i < totalSegments; i++) {
				segment = segmentList.get(i);
				segment.drawLineHairMesh(gfx);
			}
		}
		if (controls.isLineCenterVisible) {
			for (int i = startId; i < totalSegments; i++) {
				segment = segmentList.get(i);
				segment.drawLineCenter(gfx);
			}
		}

		// loadPixels();
		// mm.addFrame(pixels);
	}

	public void keyPressed() {
		if (key == ' ') {
			reset();
		} else if (key == 'c' && isControlsVisible) {
			isControlsVisible = false;
		} else if (key == 'c' && !isControlsVisible) {
			isControlsVisible = true;
		} else if (key == 'p') {
			audio.play();
		}

		if (key == 'e' || key == ESC) {
			// mm.finish(); // Finish the movie if space bar is pressed!
			exit();
		}

	}

	public void reset() {
		System.out.println("reset :: " + resetCount);

		if (resetCount == 1) {
			audio.play();
			controls.setAsStartSliders();
		} else if (resetCount == 15) {
			controls.setAsEndSliders();
		} else if (resetCount == 16) {
			// mm.finish();
			// exit();
		} else {
			controls.randomizeSliders();
		}

		resetCount++;
	}

	float[] getMouse3D() {
		((PGraphicsOpenGL) g).beginGL();
		int viewport[] = new int[4];
		double[] proj = new double[16];
		double[] model = new double[16];
		gl.glGetIntegerv(GL.GL_VIEWPORT, viewport, 0);
		gl.glGetDoublev(GL.GL_PROJECTION_MATRIX, proj, 0);
		gl.glGetDoublev(GL.GL_MODELVIEW_MATRIX, model, 0);
		FloatBuffer fb = ByteBuffer.allocateDirect(4).order(ByteOrder.nativeOrder()).asFloatBuffer();
		gl.glReadPixels(mouseX, height - mouseY, 1, 1, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT, fb);
		fb.rewind();
		double[] mousePosArr = new double[4];
		glu.gluUnProject((double) mouseX, height - (double) mouseY, (double) fb.get(0), model, 0, proj, 0, viewport, 0, mousePosArr, 0);
		((PGraphicsOpenGL) g).endGL();
		return new float[] { (float) mousePosArr[0], (float) mousePosArr[1], (float) mousePosArr[2] };
	}

	private void updateControlParams() {

		for (Segment segment : segmentList) {
			segment.segmentScalar = controls.rootScalar;
			segment.frameCountScalar = controls.rootFrameCountScalar;
			segment.seedNormalScalar = controls.rootSeedNormalScalar;

			segment.layerRadiusOffset = controls.rootLayerRadiusOffset;
			int avg = Math.abs(frameCount - segment.seed) % (totalSegments - 1);
			float layerRadiusScalar = (float) Math.sqrt(audio.getAvg(avg)) * controls.rootLayerAudioScalar + controls.rootLayerRadiusScalar;
			segment.layerRadiusScalar = layerRadiusScalar;

			segment.layerSpeedOffset = controls.rootLayerSpeedOffset;
			segment.layerRadiusOscillator = controls.rootLayerRadiusOscillator;
			segment.layerRadiusOscillatorOffset = controls.rootLayerRadiusOscillatorOffset;
			segment.flappingSpeedNormal = controls.rootFlappingSpeedNormal;
			segment.flappingSpeedOffset = controls.rootFlappingSpeedOffset;
			segment.maxAngle = controls.rootMaxAngle;
			segment.frameCount = frameCount;

		}
	}

	public void stopCamera() {
		System.out.println("stopCamera");
		// cam.setActive(false);
	}

	public void controlEvent(ControlEvent theEvent) {
		// cam.setActive(false);
	}

	public void mouseReleased() {
		cam.setActive(true);
	}

	public void setFillStyle(int colorParam, float alphaParam) {
		this.noStroke();
		this.strokeWeight(1);
		this.fill(colorParam, 255.0f * alphaParam);
	}

	public void setLineStyle(int weightParam, int colorParam, float alphaParam) {
		this.noFill();
		this.strokeWeight(weightParam);
		this.stroke(colorParam, 255.0f * alphaParam);
	}

	public float getRandom(float a, float b) {
		return (float) Math.random() * (b - a) + a;
	}

}
